/*
 * Copyright 2013-2018 Radian Memory Systems, Inc.  All Rights Reserved.
 *
 * This file is provided under non-disclosure agreement.
 * See the file COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <ctype.h>
#include <linux/types.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "nvme_express.h"
#include "nvme_ioctl.h"
typedef __u64 u64;
typedef __u32 u32;
typedef __u16 u16;
typedef __u8 u8;
#define MAX_DIES 256
#include "RMSVendorSpecific.h"
#include "nvme_util.h"
#include "dialogDef.h"
#include "rms200_hw_info.h"

#define CONTROL_DEV		"/dev/nvme0"
#define MAX_BOARD_NAME		256
#define OP_READ			1
#define OP_WRITE		2
#define PAGE_SIZE		4096

#define CRC_POLYNOMIAL          0x8005
#define CRC_INITIAL_REMAINDER   0x4F4E

/*
 * Derive parameters from the standard-specific parameters.
 */
#define CRC_WIDTH	(8 * sizeof(__u16))
#define CRC_TOPBIT	(1 << (CRC_WIDTH - 1))

/******************************************************************//**
 * @brief  Compute the CRC of a given message.
 *
 * @param message   Pointer to the start of the message to calculate CRC on.
 * @param nBytes    Length of the message in bytes to caculate CRC on.
 * @param remainder Starting remainder for CRC calculation.
 *
 * @return          This function returns the CRC of the message.
 *********************************************************************/
static __u16
crcONFI(unsigned char const message[], int nBytes, __u16 remainder)
{

	int byte;
	unsigned char bit;

	/**
	* Perform modulo-2 division, a byte at a time.
	*/
	for (byte = 0; byte < nBytes; ++byte) {
		/**
		* Bring the next byte into the remainder.
		*/
		remainder ^= (message[byte] << (CRC_WIDTH - 8));
		/**
		* Perform modulo-2 division, a bit at a time.
		*/
		for (bit = 8; bit > 0; --bit) {
			/**
			* Try to divide the current data bit.
			*/
			if (remainder & CRC_TOPBIT)
				remainder = (remainder << 1) ^ CRC_POLYNOMIAL;
			else
				remainder = (remainder << 1);
		}
	}
	/**
	* The final remainder is the CRC result.
	*/
	return remainder;

}


static void
create_rms200_dialog_cfg_cmd(
	struct usr_io *x,
	int direction,
	unsigned char *buf,
	unsigned long len)
{
	nvme_cmd_rms200_t nvme_cmd_rms200;

	memset(x, 0x0, sizeof(struct usr_io));

	x->timeout = (direction == XFER_FROM_DEV) ? 1 : 3;
	x->direction = direction;
	x->length = len;
	x->addr = (__u64) buf;

	memset(&nvme_cmd_rms200, 0x0, sizeof(nvme_cmd_rms200));

	nvme_cmd_rms200.subOpcode = (direction == XFER_FROM_DEV)
					? RMS_GET_DIALOG_CFG
					: RMS_SET_DIALOG_CFG;
	nvme_cmd_rms200.numOfDwordPrP = len/4;

	append_rms_cmd_pkt(x, &nvme_cmd_rms200);
}

static void
wait_for_q(void)
{
	system("/bin/stty raw");
	getchar();
	system("/bin/stty cooked");
}

static void
usage(void)
{
	printf("Read or write the DiaLog configuration stored in eeprom.\n");
	printf("usage:  dialog_config [devName] mode \n");
	printf("\tdevName   : The device to be used - default /dev/nvme0\n");
	printf("\tmode must be one of:\n");
	printf("\t-r <file> [-u] : Read device configuration and write into file\n");
	printf("\t            ('stdout' for writing human-readable text\n");
	printf("\t            to standard output)\n");
	printf("\t-u        : Replace spaces with underscores in parameter names.\n");
	printf("\t            Used only in read mode to create an output file\n");
	printf("\t            suitable for writing back to the device.\n");
	printf("\t-w <file> : Write device configuration from a text file\n");
}

struct HandlerFlags {
	/* This measurement is implemented */
	unsigned int implemented:1;
	/* This measurement can be negative */
	unsigned int is_signed:1;

	unsigned int reserved:30;
};

struct DataDigest {
	struct HandlerFlags	flags;
	char			*name;
	char			*units;
};

static struct DataDigest measurementNames[DIALOG_NUM_MEASUREMENTS] = {
	{ {.implemented = 1 }, "Enterprise_Flash_Controller_temperature", ""},
	{ {.implemented = 1 }, "Capacitor_1_and_2_temperature", ""},
	{ {.implemented = 1 }, "Capacitor_3_and_4_temperature", ""},
	{ {.implemented = 1 }, "Rear_vent_ambient_temperature", ""},
	{ {.implemented = 1 }, "Board_temperature", ""},
	{ {.implemented = 1 }, "Voltage_of_Capacitor_1", "(mV)"},
	{ {.implemented = 1 }, "Voltage_of_Capacitor_2", "(mV)"},
	{ {.implemented = 1 }, "Voltage_of_Capacitor_3", "(mV)"},
	{ {.implemented = 1 }, "Capacitor_pack_voltage", "(mV)"},
	{ {.implemented = 1 }, "Capacitor_pack_voltage_at_end_of_F2F", "(mV)"},
	{ {.implemented = 1, .is_signed = 1 },
		"Current_derived_from_v3-v4", "(mA)"},
	{ {.implemented = 1 }, "Derived_Energy", "(Joules)"},
	{ {.implemented = 1 }, "Derived_capacitance_of_the_pack", "(mF)"},
	{ {.implemented = 1 }, "Derived_ESR_of_capacitor_pack", "(mOhms)"},
	{ {.implemented = 1 }, "Time_to_run_F2F", "(mSec)"},
	{ {.implemented = 1 }, "Time_to_run_Restore", "(mSec)"},
	{ {.implemented = 1 }, "Time_to_charge_capacitors", "(Seconds)"},
	{ {.implemented = 1 }, "Flash_page_read_time", ""},
	{ {.implemented = 1 }, "Flash_page_write_time", ""},
	{ {.implemented = 1 }, "Flash_block_erase_time", ""},
	{ {.implemented = 1 },
	"Correctable_bits_in_error_on_reading_a_page", ""},
	{ {.implemented = 1 },
	"Correctable_bits_in_error_on_reading_the_worst_BCH_region_of_a_page",
	""},
	{ {.implemented = 1 }, "Fan_inlet_ambient_temperature", ""},
};

struct ParamDigest {
	unsigned long	offset;
	char		*name;
	char		*units;
};

#define NUM_PARAMETERS	(sizeof(parameterNames) / sizeof(struct ParamDigest))

static struct ParamDigest parameterNames[] = {
	{ offsetof(dialog_config_t, capacitor_over_heat),
		"Capacitor_over_heat", "degrees C"},
	{ offsetof(dialog_config_t, peltier_turnon),
		"Peltier_turnon", "degrees C"},
	{ offsetof(dialog_config_t, peltier_turnoff),
		"Peltier_turnoff", "degrees C"},
	{ offsetof(dialog_config_t, peltier_turnoff_cpu),
		"Peltier_turnoff_CPU", "degrees C"},
	{ offsetof(dialog_config_t, ambient_warning),
		"Ambient_temepature_warning", "degrees C"},
	{ offsetof(dialog_config_t, capacitor_temp),
		"Capacitor_temperature", "degrees C"},
	{ offsetof(dialog_config_t, cpu_overtemp_alert),
		"CPU_over_temperarure_alert", "degrees C"},
	{ offsetof(dialog_config_t, cpu_overtemp_shutdown),
		"CPU_over_temperature_shutdown", "degrees C"},
	{ offsetof(dialog_config_t, charge_voltage_max),
		"Maximum_charge_voltage", "milli Volts"},
	{ offsetof(dialog_config_t, F2F_voltage_min),
		"Minumum_F2F_voltage", "milli Volts"},
	{ offsetof(dialog_config_t, capacitor_single_voltage),
		"Capacitor_single_voltage", "milli Volts"},
	{ offsetof(dialog_config_t, capacitor_single_voltage_time),
		"Capacitor_single_voltage_time", "Seconds"},
	{ offsetof(dialog_config_t, capacitor_pack_overvoltage),
		"Capacitor_pack_over_voltage", "milli Volts"},
	{ offsetof(dialog_config_t, current_max_charging),
		"Maximum_charging_current", "milli Amps"},
	{ offsetof(dialog_config_t, current_max_discharging),
		"Maximum_discharging_current", "milli Amps"},
	{ offsetof(dialog_config_t, current_max_normal),
		"Maximum_normal_current", "milli Amps"},
	{ offsetof(dialog_config_t, F2F_max_energy),
		"Maximum_F2F_energy", "Joules"},
	{ offsetof(dialog_config_t, post_F2F_minimum_energy),
		"Post_F2F_minimum_energy", "Joules"},
	{ offsetof(dialog_config_t, minimum_farads),
		"Minimum_capacitance", "milli Farads"},
	{ offsetof(dialog_config_t, maximum_internal_resistance),
		"Maximum_internal_resistance", "milli Ohms"},
	{ offsetof(dialog_config_t, cap_pack_rated_farads),
		"Rated_capacitance", "milli"},
	{ offsetof(dialog_config_t, digital_potentiometer_value),
		"Digital_potentiometer_value", ""}
};

struct DescDigest {
	unsigned long offset;
	char *name;
};

#define NUM_DESCRIPTIONS	(sizeof(Desriptions) / sizeof(struct DescDigest))

static struct DescDigest Desriptions[] = {
		{ offsetof(dialog_config_t, format_id), "Format_id" },
		{ offsetof(dialog_config_t, version), "Version" },
		{ offsetof(dialog_config_t, description), "Description" }
};

static int scan_u8(const char *line, char *name, void *d)
{
	return sscanf(line, "%s %hhu", name, (u8 *)d);
}
static int scan_u16(const char *line, char *name, void *d)
{
	return sscanf(line, "%s %hu", name, (u16 *)d);
}
static int scan_float(const char *line, char *name, void *d)
{
	return sscanf(line, "%s %g", name, (float *)d);
}
static int scan_double(const char *line, char *name, void *d)
{
	return sscanf(line, "%s %lg", name, (double *)d);
}
static int scan_u16_hex(const char *line, char *name, void *d)
{
	return sscanf(line, "%s %hX", name, (u16 *)d);
}
static int print_float(FILE *f, void *d)
{
	return fprintf(f, "%g\n", *(float *)d);
}
static int print_double(FILE *f, void *d)
{
	return fprintf(f, "%g\n", *(double *)d);
}
static int print_u8(FILE *f, void *d)
{
	return fprintf(f, "%hhu\n", *(u8 *)d);
}
static int print_u16(FILE *f, void *d)
{
	return fprintf(f, "%hu\n", *(u16 *)d);
}
static int print_u16_hex(FILE *f, void *d)
{
	return fprintf(f, "0x%04hX\n", *(u16 *)d);
}
struct bsDigest {
	unsigned long offset;
	char *name;
	int (*scan)(const char *line, char *name, void *pData);
	int (*print)(FILE *f, void *pData);
};

#define NUM_BOARDSPECIFIC	(sizeof(BoardSpecific) / sizeof(struct bsDigest))

static struct bsDigest BoardSpecific[] = {
		{ offsetof(dialog_config_t, bs.shh.coef_a), "Steinhart_coefficient_a", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.shh.coef_b), "Steinhart_coefficient_b", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.shh.coef_c), "Steinhart_coefficient_c", &scan_float, &print_float },

		{ offsetof(dialog_config_t, bs.rlc.voltage_conversion_k), "Voltage_conversion_k", &scan_double, &print_double },
		{ offsetof(dialog_config_t, bs.rlc.current_conversion_k), "Current_conversion_k", &scan_double, &print_double },
		{ offsetof(dialog_config_t, bs.rlc.pack_to_charge_multiplier), "Pack_to_charge_multiplier", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.rlc.validCapacitanceLow), "ValidCapacitanceLow", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.rlc.validCapacitanceHigh), "ValidCapacitanceHigh", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.rlc.defaultAdjustedCapacitance), "DefaultAdjustedCapacitance", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.rlc.validEsrLow), "ValidEsrLow", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.rlc.validEsrHigh), "ValidEsrHigh", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.rlc.validEsrTyp), "ValidEsrTyp", &scan_u16, &print_u16 },

		{ offsetof(dialog_config_t, bs.jcurv.t2jCoefA), "T2jCoefA", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.jcurv.t2jCoefB), "T2jCoefB", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.jcurv.t2jCoefC), "T2jCoefC", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.jcurv.minJoules), "MinJoules", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.jcurv.maxJoules), "MaxJoules", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.jcurv.minJoulesTemp), "MinJoulesTemp", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.jcurv.maxJoulesTemp), "MaxJoulesTemp", &scan_u16, &print_u16 },

		{ offsetof(dialog_config_t, bs.ccal.cap_temp_norm), "Cap_temp_norm", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.ccal.cap_temp_slope), "Cap_temp_slope", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.ccal.cap_voltage_norm), "Cap_voltage_norm", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.ccal.cap_voltage_slope), "Cap_voltage_slope", &scan_float, &print_float },

		{ offsetof(dialog_config_t, bs.cmc.capMeasureBump), "CapMeasureBump", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.cmc.caps_min_wait_charge_interval), "Caps_min_wait_charge_interval", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.cmc.caps_wait_charge_interval), "Caps_wait_charge_interval", &scan_u16, &print_u16 },
		{ offsetof(dialog_config_t, bs.cmc.integratePercentage), "IntegratePercentage", &scan_u8, &print_u8 },
		{ offsetof(dialog_config_t, bs.cmc.chargevTrimPercentage), "ChargevTrimPercentage", &scan_u8, &print_u8 },

		{ offsetof(dialog_config_t, bs.vc.chargeInSeconds), "ChargeInSeconds", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.vc.cchg_vmin), "Cchg_vmin", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.vc.cchg_vmax), "Cchg_vmax", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.vc.default_charge_voltage), "Default_charge_voltage", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.vc.minimum_charge_voltage), "Minimum_charge_voltage", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.vc.cchg_operatingVmin), "Cchg_operatingVmin", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.vc.lowMarginTrigger), "LowMarginTrigger", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.vc.charge_hysteresis), "Charge_hysteresis", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.vc.charge_margin), "Charge_margin", &scan_float, &print_float },

		{ offsetof(dialog_config_t, bs.clife.minCapacitance), "MinCapacitance", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.clife.capElbow), "CapElbow", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.clife.cl_vmax), "Cl_vmax", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.clife.cl_vt), "Cl_vt", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.clife.cl_voff), "Cl_voff", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.clife.cl_tshift), "Cl_tshift", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.clife.cl_tdep), "Cl_tdep", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.clife.cl_rfactor), "Cl_rfactor", &scan_float, &print_float },
		{ offsetof(dialog_config_t, bs.clife.cl_lifemax), "Cl_lifemax", &scan_float, &print_float },

		{ offsetof(dialog_config_t, bs.flag1), "Flag1", &scan_u16_hex, &print_u16_hex },
		{ offsetof(dialog_config_t, bs.flag2), "Flag2", &scan_u16_hex, &print_u16_hex }
};

static char *
replace_underscores(char *str)
{
	int i;
	char *dup;

	if (str == NULL)
		return NULL;

	dup  = strdup(str);
	for (i = 0; i < strlen(dup) - 1; i++) {
		if (dup[i] == '_')
			dup[i] = ' ';
	}
	return dup;
}

static void bs_print(int bs, FILE *f, void *d)
{
	if (BoardSpecific[bs].print)
		BoardSpecific[bs].print(f, d);
}

static void
dump_dialog_config(unsigned char *buffer, FILE *fd, int underscore)
{
	int	j, p, m, t, bs;
	char   line[128];
	__u16  savedCrc;
	__u16  calculatedCrc;
	dialog_config_t	*configPtr = (dialog_config_t *) buffer;

	savedCrc = configPtr->verify_crc;
	configPtr->verify_crc = 0;
	calculatedCrc = crcONFI((__u8 *)configPtr,
				(int) sizeof(dialog_config_t),
				CRC_INITIAL_REMAINDER);
	if (calculatedCrc != savedCrc) {
		printf("CRC of the configuration is not correct.\n");
		printf("Data might be inaccurate (%04x != %04x)\n",
			calculatedCrc,
			savedCrc);
	}

	fprintf(fd, "#################\n");
	fprintf(fd, "# _Descriptions_\n");
	fprintf(fd, "#################\n");
	for (p = 0; p < NUM_DESCRIPTIONS; p++) {
		memset(line, 0, sizeof(line));
		if (underscore) {
			sprintf(line, "%s: ", Desriptions[p].name);
		} else {
			char *stripped = replace_underscores(Desriptions[p].name);
			if (stripped) {
				sprintf(line, "%s: ", stripped);
				free(stripped);
			}
		}
		fprintf(fd, "%s", line);
		for (j = strlen(line); j < 50; j++)
			fprintf(fd, " ");

		if (strcmp("Description", Desriptions[p].name))
			fprintf(fd, "%hhu\n", *(unsigned char *)(&buffer[Desriptions[p].offset]));
		else
			fprintf(fd, "\"%s\"\n", (char *)&buffer[Desriptions[p].offset]);
	}

	fprintf(fd, "###############\n");
	fprintf(fd, "# _Thresholds_\n");
	fprintf(fd, "###############\n");
	for (m = 0; m < DIALOG_NUM_MEASUREMENTS; m++) {
		memset(line, 0, sizeof(line));
		if (measurementNames[m].flags.implemented) {
			char *stripped;
			stripped = underscore
				? measurementNames[m].name
				: replace_underscores(measurementNames[m].name);
			if (stripped) {
				sprintf(line, "%s: ", stripped);
				fprintf(fd, "%s", line);
			}
			for (j = strlen(line); j < 50; j++)
				fprintf(fd, " ");
			for (t = 0; t < DIALOG_NUM_THRESHOLDS; t++) {
				fprintf(fd,
					measurementNames[m].flags.is_signed
						? "%hd "
						: "%hu ",
					configPtr->thresholds[m][t]);
			}
			fprintf(fd,
				"Exceed %d Below %d\n",
				configPtr->error_levels[m].exceed,
				configPtr->error_levels[m].below);
		}
	}
	fprintf(fd, "###############\n");
	fprintf(fd, "# _Parameters_\n");
	fprintf(fd, "###############\n");
	for (p = 0; p < NUM_PARAMETERS; p++) {
		memset(line, 0, sizeof(line));
		if (underscore) {
			sprintf(line, "%s: ", parameterNames[p].name);
		} else {
			char *stripped = replace_underscores(parameterNames[p].name);
			if (stripped) {
				sprintf(line, "%s: ", stripped);
				free(stripped);
			}
		}

		fprintf(fd, "%s", line);
		for (j = strlen(line); j < 50; j++)
			fprintf(fd, " ");
		fprintf(fd,
			"%hu (%s)\n",
			*(unsigned short *)(&buffer[parameterNames[p].offset]),
			parameterNames[p].units);
	}
	fprintf(fd, "#####################\n");
	fprintf(fd, "# _Board_ specific\n");
	fprintf(fd, "#####################\n");
	for (bs = 0; bs < NUM_BOARDSPECIFIC; bs++) {
		memset(line, 0, sizeof(line));
		if (underscore) {
			sprintf(line, "%s: ", BoardSpecific[bs].name);
		} else {
			char *stripped = replace_underscores(BoardSpecific[bs].name);
			if (stripped) {
				sprintf(line, "%s: ", stripped);
				free(stripped);
			}
		}

		fprintf(fd, "%s", line);
		for (j = strlen(line); j < 50; j++)
			fprintf(fd, " ");
		bs_print(bs, fd, &buffer[BoardSpecific[bs].offset]);
	}
}

static int
scan_dialog_thresholds(
	const char *line,
	dialog_config_t *configPtr)
{
	char comment[128];
	char names[2][128];
	unsigned short values[16];
	int sv;
	int m;
	int t;

	sv = sscanf(line,
		"%s %hd %hd %hd %hd %hd %hd %hd %hd %hd %s %hd %s %hd",
		comment,
		&values[0],
		&values[1],
		&values[2],
		&values[3],
		&values[4],
		&values[5],
		&values[6],
		&values[7],
		&values[8],
		names[0],
		&values[9],
		names[1],
		&values[10]);
	if (sv == 14) {
		/* Trim colon */
		comment[strlen(comment) - 1] = 0;
		for (m = 0; m < DIALOG_NUM_MEASUREMENTS; m++) {
			if (strcmp(comment, measurementNames[m].name))
				continue;

			for (t = 0; t < DIALOG_NUM_THRESHOLDS; t++)
				configPtr->thresholds[m][t] = values[t];

			if (strcmp(names[0], "Exceed") == 0)
				configPtr->error_levels[m].exceed = values[9];
			if (strcmp(names[1], "Below") == 0)
				configPtr->error_levels[m].below = values[10];
			break;
		}
	} else {
		fprintf(stderr,
			"Error parsing input expected 14 values, got %d\n",
			sv);
		return 0;
	}
	return 1;
}

static int
scan_dialog_parameters(
	const char *line,
	unsigned char *buffer)
{
	char comment[128];
	char names[2][128];
	unsigned short values[16];
	int sv;
	int p;
	unsigned char *ptr;

	sv = sscanf(line,
		"%s %hd %s",
		comment,
		&values[0],
		names[0]);
	if (sv == 3) {
		comment[strlen(comment) - 1] = 0;
		for (p = 0; p < NUM_PARAMETERS; p++) {
			if (strcmp(comment, parameterNames[p].name))
				continue;

			ptr = &buffer[parameterNames[p].offset];
			*(unsigned short *)ptr = values[0];
			break;
		}
	} else {
		fprintf(stderr,
			"Error parsing input expected 3 values, got %d\n",
			sv);
		return 0;
	}
	return 1;
}

static int
scan_dialog_description(
	const char *line,
	unsigned char *buffer)
{
	char comment[128];
	unsigned short value;
	char names[DIALOG_DESC_LEN_MAX] = {0};
	int sv;
	int p;
	unsigned char *ptr;
	char *start = NULL, *finish = NULL;
	int isStr = 0;
	start = strpbrk(line, "\"");
	if (start) {
		isStr = 1;
		finish = strpbrk(start+1, "\"");
	}

	if (isStr) {
		int len = finish - start -1;
		if (len > 21) {
			fprintf(stderr, "Description (%d) is being truncated to 21 characters\n", len);
			len = 21;
		}
		sv = sscanf(line, "%s", comment);
		memcpy(names, start+1, len);
		sv++;
	} else
		sv = sscanf(line, "%s %hu", comment, &value);

	if (sv == 2) {
		comment[strlen(comment) - 1] = 0;
		for (p = 0; p < NUM_DESCRIPTIONS; p++) {
			if (strcmp(comment, Desriptions[p].name))
				continue;

			ptr = &buffer[Desriptions[p].offset];
			if (isStr)
				memcpy(ptr, names, DIALOG_DESC_LEN_MAX);
			else
				*(unsigned char *)ptr = value;
			break;
		}
	} else {
		fprintf(stderr,
			"%s:%d Error parsing input expected 2 values, got %d\n", __func__, __LINE__,
			sv);
		return 0;
	}
	return 1;
}

static int bs_scan(int bs, const char *line, char *name, void *pData)
{
	if (BoardSpecific[bs].scan)
		return BoardSpecific[bs].scan(line, name, pData);
	return 0;
}

static int
scan_dialog_boardSpecific(
	const char *line,
	unsigned char *buffer)
{
	char comment[128];
	int sv;
	int bs;

	sv = sscanf(line, "%s", comment);

	if (sv == 1) {
		comment[strlen(comment) - 1] = 0;
		for (bs = 0; bs < NUM_BOARDSPECIFIC; bs++) {
			if (!strcmp(comment, BoardSpecific[bs].name))
				break;
		}
		if (bs == NUM_BOARDSPECIFIC) {
			fprintf(stderr, "No %s in BoardSpecific[]\n", comment);
			return 0;
		}
		return bs_scan(bs, line, comment, &buffer[BoardSpecific[bs].offset]);
	}
	return 0;
}

/*
 * scan a text configuration file into the structure
 */
static int
scan_dialog_config(unsigned char *buffer, char *fileName)
{
	FILE *fp;
	char *line = NULL;
	char *savedP = NULL;
	size_t len = 0;
	ssize_t readLen;
	int doThresholds = 0;
	int doParameters = 0;
	int doDescription = 0;
	int doBoard = 0;
	char comment[128];
	int ret = -1;
	dialog_config_t	*configPtr = (dialog_config_t *) buffer;

	/* Must have a file name */
	if (fileName == NULL) {
		fprintf(stderr, "Missing File name.\n");
		return ret;
	}

	fp = fopen(fileName, "r");
	if (fp == NULL) {
		perror(fileName);
		return ret;
	}

	while ((readLen = getline(&line, &len, fp)) != -1) {
		if (readLen == 1)
			continue;
		savedP = line;
		while (isspace(*line)) line++;
		if (strlen(line) == 0) {
			line = savedP;
			continue;
		}
		if (line[0] == '#') {
			if (sscanf(&line[2], "%s", comment) == 1) {
				if (strcmp(comment, "_Descriptions_") == 0) {
					doDescription = 1;
					doParameters = 0;
					doThresholds = 0;
					doBoard = 0;
				}
				if (strcmp(comment, "_Parameters_") == 0) {
					doParameters = 1;
					doThresholds = 0;
					doDescription = 0;
					doBoard = 0;
				}
				if (strcmp(comment, "_Thresholds_") == 0) {
					doThresholds = 1;
					doParameters = 0;
					doDescription = 0;
					doBoard = 0;
				}
				if (strcmp(comment, "_Board_") == 0) {
					doBoard = 1;
					doThresholds = 0;
					doParameters = 0;
					doDescription = 0;
				}
			}
			line = savedP;
			continue;
		}
		if (doDescription && !scan_dialog_description(line, buffer))
			goto done;
		if (doThresholds && !scan_dialog_thresholds(line, configPtr))
			goto done;
		if (doParameters && !scan_dialog_parameters(line, buffer))
			goto done;
		if (doBoard && !scan_dialog_boardSpecific(line, buffer))
			goto done;
		line = savedP;
	}

	ret = 0;
done:
	if (savedP) free(savedP);
	fclose(fp);

	return ret;
}


int
main(int argc, char **argv)
{
	struct usr_io x;
	unsigned long status;
	FILE *inOutFd = NULL;
	int fd, c;
	int op, flag = 0;
	char *inOutFileName = NULL;
	char *controlDev = CONTROL_DEV;
	int direction, transSize;
	unsigned char *buffer = (unsigned char *)0;
	int ret = -1;
	int underscore = 0;
	char *pEnv;

	if (argc == 1) {
		usage();
		return ret;
	}
	/* Override default device */
	if (argc >= 2 && (strncmp("/dev/", argv[1], 5) == 0)) {
		controlDev = argv[1];
	} else {
		pEnv = getenv("RMS_DEVICE");
		if (pEnv) {
            controlDev = pEnv;
		}
	}

	while (1) {

		c = getopt(argc, argv, "Hhr:uw:");

		if (c == -1)
			break;

		switch (c) {

		case 'u':
			underscore = 1;
			break;

		case 'r':
			op = OP_READ;
			if (strncmp(optarg, "stdout", 6))
				inOutFileName = optarg;
			else
				inOutFileName = NULL;
			flag++;
			break;

		case 'w':
			op = OP_WRITE;
			inOutFileName = optarg;
			flag++;
			break;

		case 'd':
			op = OP_WRITE;
			flag++;
			break;

		case 'H':
			usage();
			wait_for_q();
			exit(1);
			break;

		case 'h':
		default:
			usage();
			exit(1);
		}
	}

	if (flag > 1) {
		fprintf(stderr, "Only one of -r -w -d\n");
		return ret;
	}

	if ((flag == 0) && (errno != 0)) {
		/* No conversion could be performed */
		fprintf(stderr, "Error in argument\n");
		return ret;
	}

	fd = open(controlDev, O_RDWR);
	if (fd < 0) {
		perror(controlDev);
		return ret;
	}

	if (posix_memalign((void **) &buffer, PAGE_SIZE, PAGE_SIZE) != 0) {
		fprintf(stderr, "Memory allocation error\n");
		return ret;
	}

	/* always read from device first */
	direction = XFER_FROM_DEV;
	transSize = sizeof(dialog_config_t);
	create_rms200_dialog_cfg_cmd(&x, direction, buffer, transSize);
	status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
	if (status != 0) {
		perror("NVME_IOCTL_ADMIN_CMD");
		fprintf(stderr, "status = %ld\n", status);
		goto done;
	}
	if (x.comp.SCT != 0 || x.comp.SC != 0) {
		fprintf(stderr, "Operation failed, NVMe error = %s\n",
			nvme_err(x.comp.SCT, x.comp.SC));
		goto done;
	}

	if (op == OP_READ) {
		/* Can dump to stdout */
		if (inOutFileName == NULL) {
			inOutFd = 0;
		} else {
			inOutFd = fopen(inOutFileName, "w+");
			if (inOutFd == NULL) {
				perror(inOutFileName);
				goto done;
			}
		}
		if (inOutFd == 0)
			dump_dialog_config(buffer, stdout, underscore);
		else if (fwrite(buffer, 1, transSize, inOutFd) != transSize)
			fprintf(stderr, "Write error\n");
	}

	if (op == OP_WRITE) {
		direction = XFER_TO_DEV;
		if (scan_dialog_config(buffer, inOutFileName) < 0)
			goto done;
		/* pre-calculate the CRC, get rid of the annoying
		 * crc mismatch message when writing a config to device.
		 */
		dialog_config_t *dcp = (dialog_config_t *) buffer;
		dcp->verify_crc = 0;
		u16 calculatedCrc = crcONFI((__u8 *) dcp, (int) sizeof(dialog_config_t),
				CRC_INITIAL_REMAINDER);
		dcp->verify_crc = calculatedCrc;
		create_rms200_dialog_cfg_cmd(&x, direction, buffer, transSize);
		status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
		if (status != 0) {
			perror("NVME_IOCTL_ADMIN_CMD");
			fprintf(stderr, "status = %ld\n", status);
			goto done;
		}
		if (x.comp.SCT != 0 || x.comp.SC != 0) {
			fprintf(stderr, "Operation failed, NVMe error = %s\n",
					nvme_err(x.comp.SCT, x.comp.SC));
			goto done;
		}
	}

	ret = 0;

done:
	free(buffer);

	if (fd != 0)
		close(fd);

	if (inOutFd != NULL)
		fclose(inOutFd);
	return ret;
}
