/*
 * Copyright 2013-2018 Radian Memory Systems, Inc.  All Rights Reserved.
 *
 * This file is provided under non-disclosure agreement.
 * See the file COPYING for details.
 */
/*
 * dialog_parser.c
 *
 * Utilities to display the data collected in the dialog page
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <asm-generic/types.h>
#include <math.h>
typedef __u64 u64;
typedef __u32 u32;
typedef __u16 u16;
typedef __u8 u8;
#include "nvme_express.h"
#include "dialogDef.h"

#define NELEMENTS(array) (sizeof(array) / sizeof(*(array)))

static void
vccConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	sprintf(resultStr,
		"%.3f V",
		dialogPtr->measurements[id].recent * .00030518 + 2.5);
}

static double getChargeMultiplier(DialogPage *dialogPtr)
{
	double chargeMultiplier = 3.87;
	/*
	 * Newer models have a different resistor network
	 * to measure higher charge voltage.
	*/
	if (dialogPtr->dialogVersion >= 4) {
		int nDwords = dialogPtr->dpad.extensions.nDwords;
		if (offsetof(struct dialog_extensions, chargeMultiplier) < (nDwords + 1) * 4) {
			/* Either 3.87 or 5.32 mutiplied by 1000 */
			if (dialogPtr->dpad.extensions.chargeMultiplier != 0)
				chargeMultiplier = (double)(dialogPtr->dpad.extensions.chargeMultiplier) / 1000;
		}
	}
	return chargeMultiplier;
}

static void
voltageConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	DialogMeasurement *meas = dialogPtr->measurements;
	double vf = (double) meas[id].recent;
	double chargeMultiplier = getChargeMultiplier(dialogPtr);

	/*
	 * Voltage of Capacitor 2 relative to Capacitor 1 etc.
	 * Need to subtract other capacitors voltages
	 * Backward compatibility - the measuremnets points changed so
	 * use numbers for dialog revision 1
	 */
	if (dialogPtr->dialogVersion == 1) {
		switch (id) {
		case 5: /* DM_CAP2_VOLTAGE */
			vf = vf * 2 - meas[4].recent;
			break;
		case 6:  /* DM_CAP3_VOLTAGE */
			vf = vf * 3 - meas[5].recent * 2;
			break;
		case 7:  /* DM_CAP4_VOLTAGE */
			vf = vf * chargeMultiplier - meas[6].recent * 3;
			break;
		case DM_PACK_VOLTAGE:
		case DM_PACK_VOLT_F2F_END:
			vf = vf * chargeMultiplier;
			break;
		default:
			break;
		}
	} else {
		switch (id) {
		case DM_CAP2_VOLTAGE:
			vf = vf * 2 - meas[DM_CAP1_VOLTAGE].recent;
			break;
		case DM_CAP3_VOLTAGE:
			vf = vf * 3 - meas[DM_CAP2_VOLTAGE].recent * 2;
			break;
		case DM_CAP4_VOLTAGE:
			vf = vf * 4 - meas[DM_CAP3_VOLTAGE].recent * 3;
			break;
		case DM_PACK_VOLTAGE:
		case DM_PACK_VOLT_F2F_END:
			vf = vf * chargeMultiplier;
			break;
		default:
			break;
		}
	}
	sprintf(resultStr, "%.3f V", (vf * .00030518));
}

static double
cap4voltage(DialogPage *dialogPtr, double chargeMultiplier)
{
	double result;

	if (chargeMultiplier != 3.87)
		result = (dialogPtr->dpad.extensions.cap4recent * 4 -
				  dialogPtr->measurements[DM_CAP3_VOLTAGE].recent * 3);
	else
		result = (dialogPtr->measurements[DM_PACK_VOLTAGE].recent * chargeMultiplier -
				  dialogPtr->measurements[DM_CAP3_VOLTAGE].recent * 3);
	return result * .00030518;
}

static double
cap5voltage(DialogPage *dialogPtr, double chargeMultiplier)
{
	return (dialogPtr->measurements[DM_PACK_VOLTAGE].recent * chargeMultiplier -
		dialogPtr->dpad.extensions.cap4recent * 4) * .00030518;
}

static void
currentConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	double	vf = (double) dialogPtr->measurements[id].recent;

	/*
	 * As per LT2990 spec each bit is 19.42 uV,
	 * and our resistor is 0.01 ohm
	 */
	sprintf(resultStr, "%.3f A", vf * .001942);
}

#define THERM_RESISTOR 17400

static void
voltage2tempConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	float value, vRef, thermalRes, lnx, temperature;

	if ((id == DM_PRINCETON_TEMP) && (dialogPtr->dialogVersion > 2)) {
		sprintf(resultStr, "%d C", dialogPtr->measurements[id].recent);
		return;
	}

	/* per LTC2990 spec.: lsb = 305.18uV, and add 2.5V to vRef */
	value = dialogPtr->measurements[id].recent * .00030518;
	vRef = (dialogPtr->refVcc * .00030518) + 2.5;
	/*
	 * This calculates the thermistor resistance, based on the
	 * voltage read divided down by a 17.4K resistor
	 */
	thermalRes = THERM_RESISTOR / ((vRef / value) - 1);

	/* from resistance to temperature */
	lnx = log(thermalRes);
#define STEINHART_COEFFICIENT_A		(1.0523E-03)
#define STEINHART_COEFFICIENT_B		(2.1140E-04)
#define STEINHART_COEFFICIENT_C		(9.0821E-08)

	temperature = 1 / (STEINHART_COEFFICIENT_A
			+ STEINHART_COEFFICIENT_B * lnx
			+ STEINHART_COEFFICIENT_C * lnx * lnx * lnx) - 273.15; /* in kelvin to celsius */

	sprintf(resultStr, "%.2f C", temperature);
}

static void
capacitanceConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	sprintf(resultStr, "%d.%03d F",
			dialogPtr->measurements[id].recent / 1000,
			abs(dialogPtr->measurements[id].recent % 1000));
}

static void
resistanceConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	sprintf(resultStr, "%d.%03d Ohm",
			dialogPtr->measurements[id].recent / 1000,
			abs(dialogPtr->measurements[id].recent % 1000));
}

static void
energyConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	sprintf(resultStr, "%d Joules", dialogPtr->measurements[id].recent);
}

static void
counterConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	sprintf(resultStr, "%d", dialogPtr->measurements[id].recent);
}

static void
microTimeConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	sprintf(resultStr, "%3.2f Seconds",
			(float)(dialogPtr->measurements[id].recent) / 1000000);
}

static void
timeConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	sprintf(resultStr, "%d Seconds", dialogPtr->measurements[id].recent);
}

static void
temperatureConverter(DialogPage *dialogPtr, int id, char *resultStr)
{
	double	vf = (double) dialogPtr->measurements[id].recent;

	/* On newer boards the temperature is measured by a specific chip */
	if ((dialogPtr->dialogVersion < 4) || (getChargeMultiplier(dialogPtr) == 3.87)) {
		/* As per temperature monitor chip spec each bit is 0.0625 Degrees */
		vf *= .0625;
	}
	sprintf(resultStr, "%.2f C", vf);
}

struct EventDigest {
	DialogEventsEnum	ev;
	char				*str;
};

static struct EventDigest recentEvents[] = {
	{ DE_F2F,					"flushToFlash"},
	{ DE_PROPER_SHUTDOWN,		"cleanShutdownWithoutFlushToFlash"},
	{ DE_ERASE_ERROR,			"errorOnErase"},
	{ DE_WRITE_ERROR,			"errorOnWrite"},
	{ DE_UNCORRECTABLE_ERROR,	"uncorrectableErrorOnRead"},
	{ DE_RESTORE_INCOMPLETE,	"restoredIncompleteFlashImage"},
	{ DE_PRODUCTION,			"productionOrRepair"},
	{ DE_EXCESSIVE_CAP_VLOTAGE,	"excessiveCapacitorVoltage"},
	{ DE_EXCESSIVE_CURRENT,		"excessiveCurrent"}
};

struct HandlerFlags {
	/* This measurement is implemented */
	unsigned int implemented:1;
	/* Show bins of measurements */
	unsigned int showBins:1;
	unsigned reserved:30;
};

struct DataDigest {
	struct HandlerFlags	flags;
	char *measureName;
	void (*convertFunc)(DialogPage*, int, char*);
};

static struct DataDigest recentValHandler[DIALOG_NUM_VERSIONS][DIALOG_NUM_MEASUREMENTS] = {
	{ /* DiaLog version 1 */
		{
			{ .implemented = 1, .showBins = 1 },
			"enterpriseFlashControllerTemperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitor4Temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"ambientTemperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"rms200BoardTemperature",
			temperatureConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor1",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor2",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor3",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor4",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitorPackVoltage",
			voltageConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"capacitorPackVoltageAtEndOfFlushToFlash",
			voltageConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"currentDerivedFromV1V2",
			currentConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"derivedPower",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"derivedCapacitanceOfThePack",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"derivedSeriesResistanceOfCapacitorPack",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToRunFlushToFlash",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToRunRestore",
			microTimeConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"timeToChargeCapacitors",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashPageReadTime",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashPageWriteTime",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashBlockEraseTime",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"correctableBitsInErrorOnReadingAPage",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
	"correctableBitsInErrorOnReadingTheWorstBchRegionOfAPage",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"temperatureRefernceVoltage",
			vccConverter
		},
	},
	{ /* DiaLog version 2 */
		{
			{ .implemented = 1, .showBins = 1 },
			"enterpriseFlashControllerTemperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitor1And2Temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitor3And4Temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"ambientTemperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"rms200BoardTemperature",
			temperatureConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor1",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor2",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor3",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitorPackVoltage",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitorPackVoltageAtEndOfFlushToFlash",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"currentDerivedFromV3V4",
			currentConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedEnergy",
			energyConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedCapacitanceOfThePack",
			capacitanceConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedEsrOfCapacitorPack",
			resistanceConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToRunFlushToFlash",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToRunRestore",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToChargeCapacitors",
			timeConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashPageReadTime",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashPageWriteTime",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashBlockEraseTime",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"correctableBitsInErrorOnReadingAPage",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
"correctableBitsInErrorOnReadingTheWorstBchRegionOfAPage",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 0 },
			"temperatureRefernceVoltage",
			vccConverter
		},
	},
	{ /* DiaLog version 3 */
		{
			{ .implemented = 1, .showBins = 1 },
			"enterpriseFlashControllerTemperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitor1And2Temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitor3And4Temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"ambientTemperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"rms200BoardTemperature",
			temperatureConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor1",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor2",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor3",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitorPackVoltage",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitorPackVoltageAtEndOfFlushToFlash",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"currentDerivedFromV3V4",
			currentConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedEnergy",
			energyConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedCapacitanceOfThePack",
			capacitanceConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedEsrOfCapacitorPack",
			resistanceConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToRunFlushToFlash",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToRunRestore",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToChargeCapacitors",
			timeConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashPageReadTime",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashPageWriteTime",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashBlockEraseTime",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"correctableBitsInErrorOnReadingAPage",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
	"correctableBitsInErrorOnReadingTheWorstBchRegionOfAPage",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 0 },
			"temperatureRefernceVoltage",
			vccConverter
		},
	},
	{ /* DiaLog version 4 */
		{
			{ .implemented = 1, .showBins = 1 },
			"enterpriseFlashControllerTemperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitor1And2Temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitor3And4Temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"rearVentAmbientTemperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"rms200BoardTemperature",
			temperatureConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor1",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor2",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"voltageOfCapacitor3",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitorPackVoltage",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"capacitorPackVoltageAtEndOfFlushToFlash",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"currentDerivedFromV3V4",
			currentConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedEnergy",
			energyConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedCapacitanceOfThePack",
			capacitanceConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"derivedEsrOfCapacitorPack",
			resistanceConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToRunFlushToFlash",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToRunRestore",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"timeToChargeCapacitors",
			timeConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashPageReadTime",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashPageWriteTime",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"flashBlockEraseTime",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"correctableBitsInErrorOnReadingAPage",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
	"correctableBitsInErrorOnReadingTheWorstBchRegionOfAPage",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"fanInletAmbientTemperature",
			voltage2tempConverter
		},
	},
	{ /* DiaLog version 5 */
		{
			{ .implemented = 1, .showBins = 1 },
			"Enterprise Flash Controller temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Capacitor 1 and 2 temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Capacitor 3 and 4 temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Rear vent ambient Temperature",
			voltage2tempConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Board temperature",
			temperatureConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"Voltage of Capacitor 1",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"Voltage of Capacitor 2",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"Voltage of Capacitor 3",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Capacitor pack voltage",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Capacitor pack voltage at end of F2F",
			voltageConverter
		},
		{
			{ .implemented = 1, .showBins = 0 },
			"Current derived from v3-v4",
			currentConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Derived Energy",
			energyConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Derived capacitance of the pack",
			capacitanceConverter
		},
		{
		{ .implemented = 1, .showBins = 1 },
			"Derived ESR of capacitor pack",
			resistanceConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Time to run F2F",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Time to run Restore",
			microTimeConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Time to charge capacitors",
			timeConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"Flash page read time",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"Flash page write time",
			counterConverter
		},
		{
			{ .implemented = 0, .showBins = 1 },
			"Flash block erase time",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Correctable bits in error on reading a page",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Correctable bits in error on reading the worst BCH region of a page",
			counterConverter
		},
		{
			{ .implemented = 1, .showBins = 1 },
			"Fan inlet ambient temperature",
			voltage2tempConverter
		}
	}
};

struct ErrorExplanation {
	unsigned int  mask;
	char *name;
};

static struct ErrorExplanation errorExplanation[] = {
	{ (1 << 0), "Capacitor over-temperature" },
	{ (1 << 1), "Princeton over-temperature" },
	{ (1 << 2), "Ambient over temperature" },
	{ (1 << 3), "Individual Capacitor over-voltage" },
	{ (1 << 4), "Capacitor Pack over-voltage" },
	{ (1 << 5), "Post F2F Capacitor pack under voltage" },
	{ (1 << 6), "High Current" },
	{ (1 << 7), "Excessive power consumption" },
	{ (1 << 8), "Low Capacitor capacity" },
	{ (1 << 9), "High Internal Resistance" },
	{ (1 << 10), "Long Flush to Flash time"},
	{ (1 << 16), "Fan malfunction"}
};

struct Smart128CounterDigest {
	int  fieldOffset;
	char *fieldName;
};

static struct Smart128CounterDigest Smart128Counters[] = {
	{
		offsetof(DialogPage,
		smartData.dataUnitReadLow),
		"numberOf512ByteBlocksReadFromDdr"
	},
	{
		offsetof(DialogPage,
		smartData.dataUnitWriteLow),
		"numberOf512ByteBlocksWrittenToDdr"
	},
	{
		offsetof(DialogPage,
		smartData.hostReadCmdLow),
		"numberOfHostReadCommands"
	},
	{
		offsetof(DialogPage,
		smartData.hostWriteCmdLow),
		"numberOfHostWriteCommands"
	},
	{
		offsetof(DialogPage,
		smartData.ctrlBusyTimeLow),
		"controllerBusyTimeMinutes"
	},
	{
		offsetof(DialogPage,
		smartData.pwrCyclLow),
		"numberOfPowerCycles"
	},
	{
		offsetof(DialogPage,
		smartData.pwrOnHourLow),
		"powerOnHours"
	},
	{
		offsetof(DialogPage,
		smartData.unsaveShutdownLow),
		"unsafeShutdowns"
	},
	{
		offsetof(DialogPage,
		smartData.mediaErrLow),
		"mediaErrors"
	},
	{
		offsetof(DialogPage,
		smartData.numOfErrLogLow),
		"numberOfErrorLogs"
	}
};

/* N[0] - contains least significant bits, N[1] - most significant */
static char *
bin128ToDec(u64 *N)
{
	/* log10(x) = log2(x) / log2(10) ~= log2(x) / 3.322 */
	static char	s[128 / 3 + 1 + 1];
	u64		n[2];
	char		*p = s;
	int		i;

	memset(s, '0', sizeof(s) - 1);
	s[sizeof(s) - 1] = '\0';

	memcpy(n, N, sizeof(n));

	for (i = 0; i < 128; i++) {
		int j, carry;

		carry = (n[1] >= 0x8000000000000000);
		/* Shift n[] left, doubling it */
		n[1] = ((n[1] << 1) & 0xFFFFFFFFFFFFFFFF)
			+ (n[0] >= 0x8000000000000000);
		n[0] = ((n[0] << 1) & 0xFFFFFFFFFFFFFFFF);

		/* Add s[] to itself in decimal, doubling it */
		for (j = sizeof(s) - 2; j >= 0; j--) {
			s[j] += s[j] - '0' + carry;

			carry = (s[j] > '9');

			if (carry)
				s[j] -= 10;
		}
	}

	while ((p[0] == '0') && (p < &s[sizeof(s) - 2]))
		p++;

	return p;
}

static void
output_128(char *str, u64 *val, FILE *fd)
{
	if (val[1] == 0)
		fprintf(fd, "%s = %lld\n", str, val[0]);
	else
		fprintf(fd, "%s = %s\n", str, bin128ToDec(val));
}

static void
output_smart_health(DialogPage *x, FILE *fd)
{
	int i;
	int offset;
	unsigned char *p = (unsigned char *) x;

	fprintf(fd, "Critical Warning byte = %02x\n",
			x->smartData.criticalWarn);
	for (i = 0; i < NELEMENTS(Smart128Counters); i++) {
		offset = Smart128Counters[i].fieldOffset;
		output_128(Smart128Counters[i].fieldName,
				(u64 *)(p + offset),
				fd);
	}
}

static void
output_error_periods(DialogMeasurement *mPtr, FILE *fd)
{
	int	j, header = 1;
	time_t myTime;
	char   timeBuff[64];

	for (j = 0; j < DIALOG_WINDOW_SIZE; j++) {
		if (mPtr->span[j].start == 0)
			break;
		if (header) {
			fprintf(fd, "Error Periods:\n");
			header = 0;
		}
		myTime = (time_t) mPtr->span[j].start;
		asctime_r(localtime(&myTime), timeBuff);
		/* Clear the new line */
		timeBuff[strlen(timeBuff) - 1] = 0;
		fprintf(fd, " start: %s\n duration: %d seconds level: %d\n",
				timeBuff,
				mPtr->span[j].duration,
				mPtr->span[j].worst);
	}
}

static void
output_events_log(DialogPage *dialogPtr, FILE *fd)
{
	int	head, tail, ev, idx, base;
	int	header = 1;
	char	timeBuff[64];
	time_t	myTime;
	DialogEvent *evPtr;

	for (ev = 0; ev < DIALOG_NUM_EVENTS; ev++) {
		evPtr = &dialogPtr->events[ev];

		if (evPtr->count == 0) {
			continue;
		} else {
			if (evPtr->count < DIALOG_WINDOW_SIZE) {
				head = 0;
				base = 0;
			} else {
				head = evPtr->count % DIALOG_WINDOW_SIZE;
				base = evPtr->count - DIALOG_WINDOW_SIZE;
			}
			tail = evPtr->count % DIALOG_WINDOW_SIZE;
			if (header) {
				fprintf(fd, " Events Logs:\n");
				header = 0;
			}
			idx = 0;
			do {
				myTime = (time_t)(evPtr->log[head].time
							& 0x7fffffff);
				asctime_r(localtime(&myTime), timeBuff);
				/* Clear the new line */
				timeBuff[strlen(timeBuff) - 1] = 0;
				/*
				 * Upper bit set means time is estimated,
				 * there is no RTC
				 */
				if (evPtr->log[head].time & 0x80000000)
					strcat(timeBuff, "*");
				fprintf(fd, "%d. %s %s (%d)\n",
						base + idx,
						timeBuff,
						recentEvents[ev].str,
						evPtr->log[head].value);
				head = (head + 1) % DIALOG_WINDOW_SIZE;
				idx++;
			} while (head != tail);
		}
	}
}

static void
output_extensions(DialogPage *dialogPtr, FILE *fd)
{
	int nDwords = dialogPtr->dpad.extensions.nDwords;
	int i;

	if (nDwords <= 0)
		return;

	if (offsetof(struct dialog_extensions, capLife) >= (nDwords + 1) * 4)
		return;

	for (i = 0; i < 2; ++i) {
		if (dialogPtr->dpad.extensions.capLife[i] > 52 * 6)
			dialogPtr->dpad.extensions.capLife[i] = 52 * 6;
	}
	fprintf(fd, "-----\n");
	fprintf(fd, "Remaining capacitor life:\n");
	fprintf(fd,
		"\t%d weeks based on temperature in the past week\n",
		dialogPtr->dpad.extensions.capLife[0]);
	fprintf(fd,
		"\t%d weeks based on temperature in the past 2 months\n",
		dialogPtr->dpad.extensions.capLife[1]);

	/* That's all for now */
	return;
}

static void
output_extensions_2(DialogPage *dialogPtr, FILE *fd)
{
	int nDwords = dialogPtr->dpad.extensions.nDwords;

	if (nDwords <= 0)
		return;

	fprintf(fd,
			"Number of Fan error logs = %u\n",
			dialogPtr->dpad.extensions.nFanLogs);
}

static void
output_extra(DialogPage *dialogPtr, FILE *fd)
{
	int i;

	if (dialogPtr->dialogVersion > 1) {
		fprintf(fd, "-----\n");
		fprintf(fd, "Extra\n");
		fprintf(fd, "-----\n");
		fprintf(fd, "Initial capacitance = %d.%03d F\n",
			dialogPtr->initialCapacitance / 1000,
			dialogPtr->initialCapacitance % 1000);
		fprintf(fd, "Initial ESR = %d.%03d Ohm\n",
			dialogPtr->initialEsr / 1000,
			dialogPtr->initialEsr % 1000);
		fprintf(fd, "------------\n");
		fprintf(fd, "Board Errors\n");
		fprintf(fd, "------------\n");
		fprintf(fd, "Existing conditions = %08x\n",
			dialogPtr->errorConditionExists);
		for (i = 0; i < NELEMENTS(errorExplanation); i++) {
			if (dialogPtr->errorConditionExists
			  & errorExplanation[i].mask) {
				fprintf(fd, " - %s\n",
					errorExplanation[i].name);
			}
		}
	}
}

static inline int device_report_valid(DialogPage *dp, int m)
{
	if (dp->dialogVersion < DIALOG_NUM_VERSIONS)
		return 1;
	return dp->dpad.extensions.mflags[m/32] & (1UL << (m%32));
}
void
output_dialog(DialogPage *dialogPtr, FILE *fd)
{
	int	i, j;
	char   resultStr[32];
	char   line[128];
	time_t myTime = (time_t) dialogPtr->snapshotTime;
	int dialogVersion = dialogPtr->dialogVersion;
	struct DataDigest *handle;

	if (dialogVersion < 1 || dialogVersion > DIALOG_NUM_VERSIONS)
		return;

	fprintf(fd, "Dialog Version = %d\n", dialogPtr->dialogVersion);
	fprintf(fd, "Dialog Configuration Description = \"%s\"\n", dialogPtr->dpad.extensions.desc.description);
	fprintf(fd, "Snapshot time = %s", asctime(localtime(&myTime)));
	myTime = (time_t) dialogPtr->nextSnapshotTime;
	fprintf(fd, "Next Snapshot time = %s", asctime(localtime(&myTime)));

	for (i = 0; i < DIALOG_NUM_MEASUREMENTS; i++) {
		memset(line, 0, sizeof(line));
		memset(resultStr, 0, sizeof(resultStr));
		handle = &recentValHandler[dialogVersion - 1][i];
		if (handle->flags.implemented && device_report_valid(dialogPtr, i)) {
			/*
			 * DM_CORRECTABLE_PAGE_BITS and
			 * DM_CORRECTABLE_BCH_BITS recent measurement
			 * does not have any meaning as many pages are
			 * read during restore. Only bins are displayed
			 */
			if ((i != DM_CORRECTABLE_PAGE_BITS)
			 && (i != DM_CORRECTABLE_BCH_BITS)) {
				handle->convertFunc(dialogPtr, i, resultStr);
				sprintf(line, "%s = %s ",
					handle->measureName,
					resultStr);
			} else {
				sprintf(line, "%s ", handle->measureName);
			}
			int l, nl = strlen(line) / 50 + 1;
			for (l = 0; l < nl - 1; l++) {
				fprintf(fd, "%*.*s\n", 50, 50, &line[l * 50]);
			}
			fprintf(fd, "%-*s", 50, &line[l * 50]);
			if (handle->flags.showBins) {
				fprintf(fd, " Bins:");
				for (j = 0; j < DIALOG_NUM_LEVELS; j++) {
					fprintf(fd, " %7d",
					    dialogPtr->measurements[i].bin[j]);
				}
			}
			fprintf(fd, "\n");
			output_error_periods(&dialogPtr->measurements[i], fd);
			/*
			 * Cap 4 voltage is derived
			*/
			if (i == DM_CAP3_VOLTAGE) {
				double chargeMultiplier = getChargeMultiplier(dialogPtr);
				fprintf(fd, "Voltage of Capacitor 4 = %.3f V\n",
					cap4voltage(dialogPtr, chargeMultiplier));
				if (chargeMultiplier != 3.87) {
					fprintf(fd, "Voltage of Capacitor 5 = %.3f V\n",
					cap5voltage(dialogPtr, chargeMultiplier));
				}
			}
		}
	}
	output_extensions(dialogPtr, fd);
	output_extra(dialogPtr, fd);
	fprintf(fd, "------------\n");
	fprintf(fd, "Smart health\n");
	fprintf(fd, "------------\n");
	output_smart_health(dialogPtr, fd);
	output_extensions_2(dialogPtr, fd);
	output_events_log(dialogPtr, fd);

}

#define	XML_FILE_NAME	"dialog.xml"
#define XML_PREFIX		"?xml version=\"1.0\" encoding=\"UTF-8\"?"
#define XML_STYLE_SHEET \
		"?xml-stylesheet type=\"text/xsl\" href=\"dialog-style.xsl\"?"
#define XML_DIALOG_TAG	"dialog"

#define	XML_TAG_END_FLAG	0x1
#define	XML_NEW_LINE_FLAG	0x2

static int
fprintf_xml_tag(FILE *fd, int flags, int spaces, const char *tag)
{
	char lineBuff[256];

	if (spaces < 256) {
		if (spaces)
			memset(lineBuff, ' ', spaces);
		lineBuff[spaces] = '\0';
	}
	return fprintf(fd, "%s<%s%s>%s",
				   lineBuff,
				   (flags & XML_TAG_END_FLAG) ? "/" : "",
				   tag,
				   (flags & XML_NEW_LINE_FLAG) ? "\n" : "");
}

static int
fprintf_xml_time(FILE *fd, int seconds)
{
	time_t myTime = (time_t) seconds;
	char   timeBuff[64];

	strftime(timeBuff, 64, "%Y-%m-%d %H:%M:%S", localtime(&myTime));
	return fprintf(fd, "%s", timeBuff);
}

static void
output_error_periods_xml(DialogMeasurement *mPtr, FILE *fd, int spaces)
{
	int	j;

	for (j = 0; j < DIALOG_WINDOW_SIZE; j++) {
		if (mPtr->span[j].start == 0)
			break;
		fprintf_xml_tag(fd,
			XML_NEW_LINE_FLAG,
			spaces,
			"errorPeriod");
		spaces += 6;
		fprintf_xml_tag(fd,
			0,
			spaces,
			"startTime");
		fprintf_xml_time(fd,
			mPtr->span[j].start);
		fprintf_xml_tag(fd,
			(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
			0,
			"startTime");
		fprintf_xml_tag(fd,
			0,
			spaces,
			"duration");
		fprintf(fd,
			"%d",
			mPtr->span[j].duration);
		fprintf_xml_tag(fd,
			(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
			0,
			"duration");
		fprintf_xml_tag(fd,
			0,
			spaces,
			"worst");
		fprintf(fd,
			"%d",
			mPtr->span[j].worst);
		fprintf_xml_tag(fd,
			(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
			0,
			"worst");
		spaces -= 6;
		fprintf_xml_tag(fd,
			(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
			spaces,
			"errorPeriod");
	}
}

static void
output_events_log_xml(DialogPage *dialogPtr, FILE *fd, int spaces)
{
	int	head, tail, ev, idx;
	DialogEvent *evPtr;

	for (ev = 0; ev < DIALOG_NUM_EVENTS; ev++) {
		evPtr = &dialogPtr->events[ev];

		if (evPtr->count == 0) {
			continue;
		} else {
			fprintf_xml_tag(fd,
				XML_NEW_LINE_FLAG,
				spaces,
				"eventOccurrences");
			spaces += 6;

			fprintf_xml_tag(fd,
				0,
				spaces,
				"name");
			fprintf(fd,
				"%s",
				recentEvents[ev].str);
			fprintf_xml_tag(fd,
				(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
				0,
				"name");

			fprintf_xml_tag(fd,
				0,
				spaces,
				"count");
			fprintf(fd,
				"%d",
				evPtr->count);
			fprintf_xml_tag(fd,
				(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
				0,
				"count");

			spaces -= 6;
			fprintf_xml_tag(fd,
				(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
				spaces,
				"eventOccurrences");

			if (evPtr->count < DIALOG_WINDOW_SIZE) {
				head = 0;
			} else {
				head = evPtr->count % DIALOG_WINDOW_SIZE;
			}
			tail = evPtr->count % DIALOG_WINDOW_SIZE;
			idx = 0;
			do {
				fprintf_xml_tag(fd,
					XML_NEW_LINE_FLAG,
					spaces,
					"event");
				spaces += 6;

				fprintf_xml_tag(fd,
					0,
					spaces,
					"name");
				fprintf(fd,
					"%s",
					recentEvents[ev].str);
				fprintf_xml_tag(fd,
					(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
					0,
					"name");

				fprintf_xml_tag(fd,
					0,
					spaces,
					"time");
				fprintf_xml_time(fd,
					evPtr->log[head].time);
				fprintf_xml_tag(fd,
					(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
					0,
					"time");

				fprintf_xml_tag(fd,
					0,
					spaces,
					"value");
				fprintf(fd,
					"%d",
					evPtr->log[head].value);
				fprintf_xml_tag(fd,
					(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
					0,
					"value");

				head = (head + 1) % DIALOG_WINDOW_SIZE;
				idx++;
				spaces -= 6;
				fprintf_xml_tag(fd,
					(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
					spaces,
					"event");
			} while (head != tail);
		}
	}
}

static void
output_extra_xml(DialogPage *dialogPtr, FILE *fd, int spaces)
{
	int i;

	if (dialogPtr->dialogVersion > 1) {
		fprintf_xml_tag(fd,
			0,
			spaces,
			"initialCapacitance");
		fprintf(fd,
			"%d.%03d F",
			dialogPtr->initialCapacitance / 1000,
			dialogPtr->initialCapacitance % 1000);
		fprintf_xml_tag(fd,
			(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
			0,
			"initialCapacitance");
		fprintf_xml_tag(fd,
			0,
			spaces,
			"initialEsr");
		fprintf(fd,
			"%d.%03d Ohm",
			dialogPtr->initialEsr / 1000,
			dialogPtr->initialEsr % 1000);
		fprintf_xml_tag(fd,
			(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
			0,
			"initialEsr");

		if (dialogPtr->errorConditionExists) {
			fprintf_xml_tag(fd,
				XML_NEW_LINE_FLAG,
				spaces,
				"errorConditions");
			spaces += 6;
			for (i = 0; i < NELEMENTS(errorExplanation); i++) {
				if (dialogPtr->errorConditionExists
				  & errorExplanation[i].mask) {
					fprintf_xml_tag(fd,
						0,
						spaces,
						"name");
					fprintf(fd,
						"%s",
						errorExplanation[i].name);
					fprintf_xml_tag(fd,
						XML_TAG_END_FLAG
							| XML_NEW_LINE_FLAG,
						0,
						"name");
				}
			}
			spaces -= 6;
			fprintf_xml_tag(fd,
				(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
				spaces,
				"errorConditions");
		}
	}
}

static void
output_128_xml(char *fieldName, u64 *val, FILE *fd, int spaces)
{
	fprintf_xml_tag(fd,
		XML_NEW_LINE_FLAG,
		spaces,
		"smartCounters");
	spaces += 6;
	fprintf_xml_tag(fd,
		0,
		spaces,
		"name");
	fprintf(fd,
		"%s",
		fieldName);
	fprintf_xml_tag(fd,
		(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
		0,
		"name");

	fprintf_xml_tag(fd,
		0,
		spaces,
		"value");
	if (val[1] == 0) {
		fprintf(fd,
			"%lld",
			val[0]);
	} else {
		fprintf(fd,
			"%s",
			bin128ToDec(val));
	}
	fprintf_xml_tag(fd,
		(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
		0,
		"value");
	spaces -= 6;
	fprintf_xml_tag(fd,
		(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
		spaces,
		"smartCounters");
}

static void
output_smart_health_xml(DialogPage *x, FILE *fd, int spaces)
{
	int i;
	int offset;
	unsigned char *p = (unsigned char *) x;

	for (i = 0; i < NELEMENTS(Smart128Counters); i++) {

		offset = Smart128Counters[i].fieldOffset;
		output_128_xml(Smart128Counters[i].fieldName,
			(u64 *)(p + offset),
			fd,
			spaces);
	}
}

void
output_dialog_xml(DialogPage *dialogPtr, FILE *inFd)
{
	int	i, j;
	char	resultStr[32];
	char	line[128];
	time_t	myTime = (time_t) dialogPtr->snapshotTime;
	char	timeBuff[64];
	int	spaces = 0;
	char	level[16];
	FILE	*fd;
	int	dialogVersion = dialogPtr->dialogVersion;
	struct DataDigest *handle;
	DialogMeasurement *meas;

	if (dialogVersion < 1 || dialogVersion > DIALOG_NUM_VERSIONS)
		return;

	if (inFd == NULL) {
		fd = fopen(XML_FILE_NAME, "w+");
		if (fd < 0) {
			printf("Could not open file %s ", XML_FILE_NAME);
			return;
		}
	} else {
		fd = inFd;
	}

	fprintf_xml_tag(fd,
		XML_NEW_LINE_FLAG,
		spaces,
		XML_PREFIX);
	fprintf_xml_tag(fd,
		XML_NEW_LINE_FLAG,
		spaces,
		XML_STYLE_SHEET);
	spaces += 3;
	fprintf_xml_tag(fd,
		XML_NEW_LINE_FLAG,
		spaces,
		XML_DIALOG_TAG);
	spaces += 6;

	fprintf_xml_tag(fd,
		0,
		spaces,
		"dialogVersion");
	fprintf(fd,
		"%d",
		dialogPtr->dialogVersion);
	fprintf_xml_tag(fd,
		(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
		0,
		"dialogVersion");
	fprintf_xml_tag(fd,
		0,
		spaces,
		"description");
	fprintf(fd,
		"\"%s\"",
		dialogPtr->dpad.extensions.desc.description);
	fprintf_xml_tag(fd,
		(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
		0,
		"description");
	fprintf_xml_tag(fd,
		0,
		spaces,
		"snapshotTime");
	strftime(timeBuff, 64, "%Y-%m-%d %H:%M:%S", localtime(&myTime));
	fprintf(fd,
		"%s",
		timeBuff);
	fprintf_xml_tag(fd,
		(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
		0,
		"snapshotTime");

	for (i = 0; i < DIALOG_NUM_MEASUREMENTS; i++) {
		memset(line, 0, sizeof(line));
		memset(resultStr, 0, sizeof(resultStr));
		handle = &recentValHandler[dialogVersion - 1][i];
		meas = &dialogPtr->measurements[i];
		if (handle->flags.implemented) {
			fprintf_xml_tag(fd,
				XML_NEW_LINE_FLAG,
				spaces,
				"measurement");
			spaces += 6;
			fprintf_xml_tag(fd,
				0,
				spaces,
				"name");
			fprintf(fd,
				"%s",
				handle->measureName);
			fprintf_xml_tag(fd,
				(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
				0,
				"name");
			if ((i != DM_CORRECTABLE_PAGE_BITS)
			 && (i != DM_CORRECTABLE_BCH_BITS)) {
				fprintf_xml_tag(fd,
					0,
					spaces,
					"recent");
				handle->convertFunc(dialogPtr, i, resultStr);
				fprintf(fd,
					"%s",
					resultStr);
				fprintf_xml_tag(fd,
					(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
					0,
					"recent");
			}

			if (handle->flags.showBins) {
				for (j = 0; j < DIALOG_NUM_LEVELS; j++) {
					sprintf(level, "level_%d", j);
					fprintf_xml_tag(fd,
						0,
						spaces,
						level);
					fprintf(fd,
						" %d",
						meas->bin[j]);
					fprintf_xml_tag(fd,
						XML_TAG_END_FLAG
							| XML_NEW_LINE_FLAG,
						0,
						level);
				}
			}

			output_error_periods_xml(meas, fd, spaces);
			spaces -= 6;
			fprintf_xml_tag(fd,
				(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
				spaces,
				"measurement");
			/*
			 * Cap 4 voltage is derived
			*/
			if (i == DM_CAP3_VOLTAGE) {
				double chargeMultiplier = getChargeMultiplier(dialogPtr);

				handle->convertFunc(dialogPtr, i, resultStr);

				fprintf_xml_tag(fd,
					XML_NEW_LINE_FLAG,
					spaces,
					"measurement");
				spaces += 6;
				fprintf_xml_tag(fd,
					0,
					spaces,
					"name");
				fprintf(fd,
					"%s",
					"voltageOfCapacitor4");
				fprintf_xml_tag(fd,
					(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
					0,
					"name");
				fprintf_xml_tag(fd,
					0,
					spaces,
					"recent");
				fprintf(fd,
					"%.3f V",
					cap4voltage(dialogPtr, chargeMultiplier));
				fprintf_xml_tag(fd,
					(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
					0,
					"recent");
				spaces -= 6;
				fprintf_xml_tag(fd,
					(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
					spaces,
					"measurement");

				if (chargeMultiplier != 3.87) {
					fprintf_xml_tag(fd,
						XML_NEW_LINE_FLAG,
						spaces,
						"measurement");
					spaces += 6;
					fprintf_xml_tag(fd,
						0,
						spaces,
						"name");
					fprintf(fd,
						"%s",
						"voltageOfCapacitor5");
					fprintf_xml_tag(fd,
						(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
						0,
						"name");
					fprintf_xml_tag(fd,
						0,
						spaces,
						"recent");
					fprintf(fd,
						"%.3f V",
						cap5voltage(dialogPtr, chargeMultiplier));
					fprintf_xml_tag(fd,
						(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
						0,
						"recent");
					spaces -= 6;
					fprintf_xml_tag(fd,
						(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
						spaces,
						"measurement");
				}
			}
		}
	}
	output_events_log_xml(dialogPtr, fd, spaces);
	output_extra_xml(dialogPtr, fd, spaces);
	output_smart_health_xml(dialogPtr, fd, spaces);
	spaces -= 6;
	fprintf_xml_tag(fd,
		(XML_TAG_END_FLAG | XML_NEW_LINE_FLAG),
		spaces,
		XML_DIALOG_TAG);

	if (inFd == NULL)
		fclose(fd);
}

/*
 * This function strips a string of a character
 */
static char *
strip(char *dst, const char *src, const char target, const int sz)
{
	char	*ret = dst;
	char	ch;
	int	i;

	for (i = 0; i < sz; i++) {
		ch = *src++;
		if (ch == '\0')
			break;
		else if (ch != target)
			*dst++ = ch;
	}
	*dst = '\0';

	return ret;
}

/*
 * This function outputs board temperatures
 */
void
output_dialog_temperatures(DialogPage *dialogPtr, FILE *fd)
{
	static const DialogMeasurementsEnum temperatureIndex[] = {
		DM_PRINCETON_TEMP,		/* Temp of the Controller */
		DM_FAN_INLET_AMBIENT_TEMP,	/* Inlet Temp */
		DM_AMBIENT_TEMP,		/* Ambient Temp */
		DM_CAP12_TEMP,			/* Temp of Capacitors 1 & 2 */
		DM_CAP34_TEMP,			/* Temp of Capacitors 3 & 4 */
		DM_BOARD_TEMP,			/* Temp of RMS 200 board */
	};
	const int NOF_TEMPERATURES = NELEMENTS(temperatureIndex);
	int i, j;
	char resultStr[32];
	int dialogVersion = dialogPtr->dialogVersion;
	char *sn;
	struct DataDigest *handle;

	sn = getenv("RMS_SN");
	if (sn) {
		memset(resultStr, 0, sizeof(resultStr));
		strip(resultStr, sn, ' ', sizeof(resultStr) - 1);
		fprintf(fd, "%8s ", sn);
	}

	for (j = 0; j < NOF_TEMPERATURES; j++) {
		memset(resultStr, 0, sizeof(resultStr));
		i = (int)temperatureIndex[j];
		handle = &recentValHandler[dialogVersion - 1][i];

		if (dialogVersion < 4 || !handle->flags.implemented) {
			sprintf(resultStr, "--");
		} else {
			handle->convertFunc(dialogPtr, i, resultStr);
			strip(resultStr, resultStr, ' ', sizeof(resultStr) - 1);
			strip(resultStr, resultStr, 'C', sizeof(resultStr) - 1);
		}
		fprintf(fd, " %10s ", resultStr);
	}
	fprintf(fd, "\n");
}

void
output_dialog_date(DialogPage *dialogPtr, int count, FILE *fd)
{
	time_t myTime = (time_t) dialogPtr->snapshotTime;
	int dialogVersion = dialogPtr->dialogVersion;

	if (dialogVersion < 1 || dialogVersion > DIALOG_NUM_VERSIONS)
		return;

	fprintf(fd,
		"%d.  Snapshot time = %s",
		count,
		asctime(localtime(&myTime)));
}

void
output_dialog_power_on(DialogPage *dialogPtr, FILE *fd)
{
	int offset;
	unsigned char *p = (unsigned char *) dialogPtr;

	offset = Smart128Counters[6].fieldOffset;
	output_128(Smart128Counters[6].fieldName, (u64 *)(p + offset), fd);
}
