/*
 * Copyright 2013-2018 Radian Memory Systems, Inc.  All Rights Reserved.
 *
 * This file is provided under non-disclosure agreement.
 * See the file COPYING for details.
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include "nvme_express.h"
#include "nvme_ioctl.h"
typedef __u64 u64;
typedef __u32 u32;
typedef __u16 u16;
typedef __u8 u8;
#define MAX_DIES 256
#include "RMSVendorSpecific.h"
#include "nvme_util.h"

#define CONTROL_DEV		"/dev/nvme0"
#define XFER_BUF (4 * 1024)

#define FW_STORE 0
#define FW_STORE_ACTIVATE 1
#define FW_ACTIVATE 2

static void
create_download_cmd(
	struct usr_io *x,
	unsigned char *buf,
	unsigned long len,
	unsigned long offset)
{
	memset(x, 0x0, sizeof(struct usr_io));
	x->cmd.header.opCode = NVM_ADMIN_CMD_FIRMWARE_DOWNLOAD;
	x->direction = XFER_TO_DEV;
	x->timeout = 1;
	x->length = len;
	x->addr = (__u64) buf;
	/*
	 * For backward compatibility with RMS-200 FW
	 * leave the number of words as 1 based.
	 * Other platforms can handle both ways for
	 * this command.
	 */
	x->cmd.cmd.firmwareDownload.numDW = (len >> 2);
	x->cmd.cmd.firmwareDownload.offset = offset >> 2;
}

static void
create_activate_cmd(
	struct usr_io *x,
	int action,
	int slot)
{
	memset(x, 0x0, sizeof(struct usr_io));
	x->cmd.header.opCode = NVM_ADMIN_CMD_FIRMWARE_ACTIVATE;
	x->direction = XFER_TO_DEV;
	x->timeout = 3;
	x->cmd.cmd.firmwareActivate.action = action;
	x->cmd.cmd.firmwareActivate.slot = slot;
}

static void
wait_for_q(void)
{
	system("/bin/stty raw");
	getchar();
	system("/bin/stty cooked");
}

static void
description(void)
{
	printf("Upload firmware to the device.  The uploaded version\n");
	printf("of the firmware will be active upon board reset, unless\n");
	printf("the -w flag is specified.  Firmware can be downloaded\n");
	printf("to slots 1 - 4.  Conventionally the  \"admin\" slot is 1\n");
	printf("and the \"user\" slot is 2.\n");
}

static void
usage(void)
{
	printf("usage: firmware_update [devName] mode [option] [<filename>]\n");
	printf("   Write firmware image <filename> to device's flash\n");
	printf("devName: The device to be updated - default /dev/nvme0\n");
	printf("mode is one of\n");
	printf("   -u(--user)         user firmware slot (2)\n");
	printf("   -s(--slot=) <slot> explicit firmware slot (2 - 4)\n");
	printf("option may be\n");
	printf("   -w(--write-only)   write new image but don't activate it\n");
	printf("<filename> is \"firmware.tar\" if omitted\n");
}

int
main(int argc, char **argv)
{
	struct usr_io x;
	unsigned char *buffer = (unsigned char *) 0;
	int fd = 0;
	int firmwareFile = 0;
	int status;
	int nbytes;
	int totalOut;
	int c;
	int slot = 0;
	int action = FW_STORE_ACTIVATE;
	char *control_dev = CONTROL_DEV;
	int aux_arg = 0;
	char *filename;

	int i = posix_memalign((void **) &buffer, XFER_BUF, XFER_BUF);

	if (i != 0) {
		printf("Memory allocation error\n");
		exit(1);
	}

	if (argc == 1) {
		usage();
		goto done;
	}

	/* Override default device */
	if (strncmp("/dev/", argv[1], 5) == 0) {
		control_dev = argv[1];
		aux_arg++;
	}

	while (1) {
		int option_index = 0;

		static struct option long_options[] = {
			{"user", 0, 0, 'u'},
			{"slot", 1, 0, 's'},
			{"write-only", 0, 0, 'w'},
			{"help", 0, 0, 'h'},
			{"Help", 0, 0, 'H'},
			{0, 0, 0, 0}
		};

		c = getopt_long(argc,
				argv,
				"us:whH",
				long_options,
				&option_index);

		if (c == -1)
			break;

		switch (c) {

		case 'u':
			if (slot != 0) {
				printf("error: use only one of -s or -u!\n");
				exit(1);
			}
			slot = 2;
			break;

		case 's':
			if (slot != 0) {
				printf("error: use only one of -s or -u!\n");
				exit(1);
			}
			slot = atoi(optarg);
			if (slot < 2 || slot > 4) {
				printf("error: slot %d invalid\n", slot);
				printf("slot must be 2 through 4\n");
				exit(1);
			}
			break;

		case 'w':
			action = FW_STORE;
			break;

		case 'H':
			description();
			usage();
			wait_for_q();
			exit(1);
			break;

		case 'h':
		default:
			usage();
			exit(1);
		}
	}

	if (slot == 0) {
		usage();
		exit(1);
	}

	optind += aux_arg;
	if (optind == argc - 1)
		filename = argv[optind];
	else if (optind == argc)
		filename = "firmware.tar";
	else {
		usage();
		exit(1);
	}

	if (strstr(filename, "firmware") == NULL) {
		printf("error: Filename must contain the string 'firmware'\n");
		exit(1);
	}

	printf("write file %s to slot %d\n", filename, slot);
	if (action == FW_STORE)
		printf("update, but don't activate...\n");

	if (posix_memalign((void **) &buffer, XFER_BUF, XFER_BUF) != 0) {
		printf("Memory allocation error\n");
		exit(1);
	}

	fd = open(control_dev, O_RDWR);
	if (fd < 0) {
		printf("Could not open file %s ", control_dev);
		if (errno == EACCES)
			printf(" Access error - try running as su\n");
		else
			printf(" Error %d\n", errno);
		exit(1);
	}

	firmwareFile = open(filename, O_RDONLY);
	if (firmwareFile < 0) {
		printf("Cant open %s\n", filename);
		exit(1);
	}

	totalOut = 0;

	while ((nbytes = read(firmwareFile, buffer, XFER_BUF)) == XFER_BUF) {
		create_download_cmd(&x, buffer, XFER_BUF, totalOut);
		status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
		if (status != 0) {
			printf("status = %d\n", status);
			goto done;
		}
		if (x.comp.SCT != 0 || x.comp.SC != 0) {
			printf("Download operation failed, NVMe error = %s\n",
				nvme_err(x.comp.SCT, x.comp.SC));
			goto done;
		}
		totalOut += XFER_BUF;
	}
	if (nbytes > 0) {
		create_download_cmd(&x, buffer, nbytes, totalOut);
		status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
		if (status != 0) {
			printf("status = %d\n", status);
			goto done;
		}
		if (x.comp.SCT != 0 || x.comp.SC != 0) {
			printf("Download operation failed, NVMe error = %s\n",
				nvme_err(x.comp.SCT, x.comp.SC));
			goto done;
		}
		totalOut += nbytes;
	}

	printf("%d bytes written to target\n", totalOut);
	create_activate_cmd(&x, action, slot);
	if (action != FW_STORE)
		printf("activate!\n");
	else
		printf("store!\n");
	status = ioctl(fd, NVME_IOCTL_ADMIN_CMD, &x);
	printf("status = %d\n", status);
	if (x.comp.SCT != 0 || x.comp.SC != 0) {
		/*
		* For the following combination - it's not realy an error
		*/
		if (x.comp.SCT == 1 && x.comp.SC == 0x0b) {
			printf("%s\n",
				nvme_err(x.comp.SCT,
				x.comp.SC));
		} else {
			printf("Activate operation failed, NVMe error = %s\n",
				nvme_err(x.comp.SCT,
				x.comp.SC));
		}
	}

done:
	if (fd > 0)
		close(fd);
	if (firmwareFile > 0)
		close(firmwareFile);
	if (buffer)
		free(buffer);
	exit(0);
}
